# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Read the detailed results of a thermal mapping analysis from a folder of CSV
files output by a thermal mapping component.
_
Detailed results include temperature amd thermal condition results. It also
includes metrics that give a sense of how hot or cold condition are like
pmv, utci category, or adaptive comfort degrees from neutral temperature.

-
    Args:
        _comf_result: Path to a folder containing CSV files output by a thermal
            mapping component.
        _load: Set to True to load the data from the CSV files into Grasshopper.

    Returns:
        comf_mtx: A Matrix object that can be connected to the "HB Visualize Thermal
            Map" component in order to spatially visualize results. This Matrix
            object can also be connected to the "LB Deconstruct Matrix"
            component to obtain detailed point-by-point and hour-by-hour
            values.
            _
            When deconstructed, each sub-list of the matrix (aka. branch of the
            Data Tree) represents one of the sensor grids used for analysis.
            The length of each sub-list matches the number of points in the
            grid. Each value in the sub-list is an hourly data collection
            containing hour-by-hour results for each point.
"""

ghenv.Component.Name = 'HB Read Thermal Matrix'
ghenv.Component.NickName = 'ThermalMtx'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '7 :: Thermal Map'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import os
import json

try:
    from ladybug.header import Header
    from ladybug.datacollection import HourlyContinuousCollection, \
        HourlyDiscontinuousCollection
    from ladybug.futil import csv_to_num_matrix
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _load:
    # parse the result_info.json into a data collection header
    with open(os.path.join(_comf_result, 'results_info.json')) as json_file:
        data_header = Header.from_dict(json.load(json_file))
    a_per = data_header.analysis_period
    continuous = True if a_per.st_hour == 0 and a_per.end_hour == 23 else False
    if not continuous:
        dates = a_per.datetimes

    # parse the grids_info.json with the correct order of the grid files
    with open(os.path.join(_comf_result, 'grids_info.json')) as json_file:
        grid_list = json.load(json_file)

    # loop through the grid CSV files, parse their results, and build data collections
    comf_matrix = []
    for grid in grid_list:
        grid_name = grid['full_id'] if 'full_id' in grid else 'id'
        metadata = {'grid': grid_name}
        grid_file = os.path.join(_comf_result, '{}.csv'.format(grid_name))
        data_matrix = csv_to_num_matrix(grid_file)
        grid_data = []
        for i, row in enumerate(data_matrix):
            header = data_header.duplicate()
            header.metadata = metadata.copy()
            header.metadata['sensor_index'] = i
            data = HourlyContinuousCollection(header, row) if continuous else \
                HourlyDiscontinuousCollection(header, row, dates)
            grid_data.append(data)
        comf_matrix.append(grid_data)

    # wrap the maptrix into an object so that it does not slow the Grasshopper UI
    comf_mtx = objectify_output(
        '{} Matrix'.format(data_header.data_type.name), comf_matrix)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply a detailed Ironbug HVAC to Honeybee Rooms or a Honeybee Model.
-

    Args:
        _hb_objs: Honeybee Rooms to which the input Ironbug HVAC will be assigned.
            This can also be a Honeybee Model for which the relevant Rooms
            referenced in the _hvac_system will be assigned the HVAC system.
        _hvac_system: A fully-detailed Irongbug HVAC system to be assigned to the
            input rooms (or model). 
        _name_: Text to set the name for the HVAC system and to be incorporated into
            unique HVAC identifier. If the name is not provided, a random name
            will be assigned.

    Returns:
        report: Reports, errors, warnings, etc.
        hb_objs: The input Rooms or Model with the detailed HVAC system applied.
"""

ghenv.Component.Name = "HB Detailed HVAC"
ghenv.Component.NickName = 'DetailedHVAC'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '4 :: HVAC'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

import json

try:  # import the honeybee extension
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
    from honeybee.model import Model
    from honeybee.room import Room
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy extension
    from honeybee_energy.config import folders
    from honeybee_energy.hvac.detailed import DetailedHVAC
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

if folders.ironbug_exe is None:
    msg = 'An installation of Ironbug that is compatible with this component\n' \
        'was not found on this machine. This component will not be usable.'
    print(msg)
    give_warning(ghenv.Component, msg)
elif folders.ironbug_version is not None:
    if folders.ironbug_version < (1, 9, 1):
        msg = 'Ironbug version "{}" is not compatible with this component.\n' \
            'This component will not be usable.'.format(
                '.'.join([str(i) for i in folders.ironbug_version]))
        print(msg)
        give_warning(ghenv.Component, msg)
else:
    msg = 'An installation of Ironbug was found at: {}\nbut it is not ' \
        'accessible. Contact your adminstrator.'.format(folders.ironbug_exe)
    print(msg)
    give_warning(ghenv.Component, msg)


if all_required_inputs(ghenv.Component):
    # extract any rooms from input Models and duplicate the rooms
    rooms, hb_objs = [], []
    for hb_obj in _hb_objs:
        if isinstance(hb_obj, Model):
            new_obj = hb_obj.duplicate()
            hb_objs.append(new_obj)
            rooms.extend(new_obj.rooms)
        elif isinstance(hb_obj, Room):
            new_obj = hb_obj.duplicate()
            hb_objs.append(new_obj)
            rooms.append(new_obj)
        else:
            raise ValueError(
                'Expected Honeybee Room or Model. Got {}.'.format(type(hb_obj)))

    # create the HVAC
    name = clean_and_id_ep_string('Detailed HVAC') if _name_ is None else \
        clean_ep_string(_name_)
    specification = json.loads(_hvac_system.ToJson())
    hvac = DetailedHVAC(name, specification)
    if _name_ is not None:
        hvac.display_name = _name_

    # apply the HVAC system to the rooms
    hvac_rooms = set(hvac.thermal_zones)
    hvac_count, rel_rooms, no_setp_rooms = 0, [], []
    for room in rooms:
        if room.identifier in hvac_rooms:
            room.properties.energy.hvac = hvac
            rel_rooms.append(room.identifier)
            hvac_count += 1
            if room.properties.energy.setpoint is None:
                no_setp_rooms.append(room.full_id)

    # give a warning if no rooms were assigned the HVAC or if there are missing rooms
    if hvac_count == 0:
        msg = 'None of the connected Rooms are referenced under the Ironbug HVAC system.\n' \
            'Make sure that the system has been set up with the correct Rooms.'
        print(msg)
        give_warning(ghenv.Component, msg)
    if len(rel_rooms) != len(hvac_rooms):
        missing_rooms = []
        found_rooms = set(rel_rooms)
        for rm_id in hvac_rooms:
            if rm_id not in found_rooms:
                missing_rooms.append(rm_id)
        msg = 'The Ironbug HVAC system contains the following rooms that are not ' \
            'in the connected _hb_objs.\n{}'.format('\n'.join(missing_rooms))
        print(msg)
        give_warning(ghenv.Component, msg)
    if len(no_setp_rooms) != 0:
        msg = 'The following Rooms have the HVAC system assigned to them '\
        'but they lack a thermostat setpoint specification.\nSetpoints ' \
        'must be assigned to these Rooms in order to be simulate-able.\n{}'.format(
            '\n'.join(no_setp_rooms))
        print(msg)
        give_warning(ghenv.Component, msg)


# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a vegetation material representing both plants and soil, which can be
plugged into the "HB Opaque Construction" component.
_
Note that the underlying models for this material were developed using horizontal
roofs and caution should be taken when applying ito to vertical walls.
-

    Args:
        _name_: Text to set the name for the material and to be incorporated into
            a unique material identifier.
        _plant_height_: A number between 0.005 and 1.0 for the height of plants in the
            vegetation layer [m]. (Default: 0.2 m).
        _leaf_area_ind_: A number between 0.001 and 5.0 for the projected leaf area per unit
            area of soil surface (aka. Leaf Area Index or LAI). Note that
            the fraction of vegetation cover is calculated directly from LAI
            using an empirical relation. (Default: 1.0).
        _leaf_reflect_: A number between 0.05 and 0.5 for the fraction of incident solar
            radiation that is reflected by the leaf surfaces. Solar radiation
            includes the visible spectrum as well as infrared and ultraviolet
            wavelengths. Typical values are 0.18 to 0.25. (Default: 0.22).
        _leaf_emiss_: A number between 0.8 and 1.0 for the ratio of thermal radiation
            emitted from leaf surfaces to that emitted by an ideal black
            body at the same temperature. (Default: 0.95).
        _soil_reflect_: A number between 0 and 1 for the fraction of incident solar
            radiation reflected by the soil material. (Default: 0.3).
        _soil_emiss_: A number between 0 and 1 for the fraction of incident long wavelength
            radiation that is absorbed by the soil material. (Default: 0.9).
        _stomat_resist_: A number between 50 and 300 for the resistance of the plants
            to moisture transport [s/m]. Plants with low values of stomatal resistance
            will result in higher evapotranspiration rates than plants with high
            resistance. (Default: 180).
        _thickness_: Number for the thickness of the soil layer [m]. (Default: 0.1).
        _conductivity_: Number for the thermal conductivity of the soil [W/m-K]. (Default: 0.35).
        _density_: Number for the density of the soil [kg/m3]. (Default: 1100).
        _spec_heat_: Number for the specific heat of the soil [J/kg-K]. (Default: 1200).

    Returns:
        mat: A vegetation material that can be assigned to a Honeybee
            Opaque construction.
"""

ghenv.Component.Name = 'HB Vegetation Material'
ghenv.Component.NickName = 'VegetationMat'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.material.opaque import EnergyMaterialVegetation
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


# set the default material properties
_plant_height_ = 0.2 if _plant_height_ is None else _plant_height_
_leaf_area_ind_ = 1.0 if _leaf_area_ind_ is None else _leaf_area_ind_
_leaf_reflect_ = 0.22 if _leaf_reflect_ is None else _leaf_reflect_
_leaf_emiss_ = 0.95 if _leaf_emiss_ is None else _leaf_emiss_
soil_abs = 0.7 if _soil_reflect_ is None else 1 - _soil_reflect_
_soil_emiss_ = 0.9 if _soil_emiss_ is None else _soil_emiss_
_stomat_resist_ = 180 if _stomat_resist_ is None else _stomat_resist_
_thickness_ = 0.1 if _thickness_ is None else _thickness_
_conductivity_ = 0.35 if _conductivity_ is None else _conductivity_
_density_ = 1100 if _density_ is None else _density_
_spec_heat_ = 1200 if _spec_heat_ is None else _spec_heat_
name = clean_and_id_ep_string('VegetationMaterial') if _name_ is None else \
    clean_ep_string(_name_)

# create the material
mat = EnergyMaterialVegetation(
    name, _thickness_, _conductivity_, _density_, _spec_heat_, 'MediumRough',
    _soil_emiss_, soil_abs, None, _plant_height_, _leaf_area_ind_,
    _leaf_reflect_, _leaf_emiss_, _stomat_resist_
)
if _name_ is not None:
    mat.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Search for available ProgramTypes within the honeybee energy standards library.
_
Note that the Room ProgramTypes output from this component effectively map to
space types within OpenStudio.
-

    Args:
        bldg_prog_: Text for the building program to search (eg. "LargeOffice",
            "MidriseApartment", etc.). The Honeybee "Building Programs" component
            lists all of the building programs available in the library. If None,
            all ProgramTypes within the library will be output (filtered by
            keywords_ below).
        _vintage_: Text for the building vintage to search (eg. "2019", "pre_1980",
            etc.). The Honeybee "Building Vintages" component lists all of the
            vintages available in the library. Default: "2019" (for ASHRAE 90.1
            2019 | IECC 2015). Note that vintages are often called "templates"
            within the OpenStudio standards gem and so this property effective
            maps to the standards gem "template".
        keywords_: Optional keywords to be used to narrow down the output list of
            room programs. If nothing is input here, all available room programs
            will be output.

    Returns:
        room_prog: A list of room program identifiers that meet the input criteria and
            can be applied to Honeybee Rooms.
"""

ghenv.Component.Name = "HB Search Programs"
ghenv.Component.NickName = 'SearchProg'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = "2"

try:
    from honeybee.search import filter_array_by_keywords
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.lib.programtypes import STANDARDS_REGISTRY
    from honeybee_energy.lib.programtypes import PROGRAM_TYPES
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


if bldg_prog_ is not None:
    # set the default vintage
    _vintage_ = _vintage_ if _vintage_ is not None else '2019'
    try:  # get the available programs for the vintage
        vintage_subset = STANDARDS_REGISTRY[_vintage_]
    except KeyError:
        raise ValueError(
            'Input _vintage_ "{}" is not valid. Choose from:\n'
            '{}'.format(_vintage_, '\n'.join(STANDARDS_REGISTRY.keys())))
    try:  # get the available programs for the building type
        room_programs = vintage_subset[bldg_prog_]
    except KeyError:
        raise ValueError(
            'Input bldg_prog_ "{}" is not avaible for vintage "{}". Choose from:\n'
             '{}'.format(bldg_prog_, _vintage_, '\n'.join(vintage_subset.keys())))
    # apply any keywords
    if keywords_ != []:
        room_programs = filter_array_by_keywords(room_programs, keywords_, False)
    # join vintage, building program and room programs into a complete string
    room_prog = ['{}::{}::{}'.format(_vintage_, bldg_prog_, rp) for rp in room_programs]
else:
    # return all programs in the library filtered by keyword
    room_prog = sorted(PROGRAM_TYPES)
    if _vintage_ is not None:
        room_prog = filter_array_by_keywords(room_prog, [_vintage_], False)
    if keywords_ != []:
        room_prog = filter_array_by_keywords(room_prog, keywords_, False)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize face and sub-face level energy simulation results as colored geometry.

-
    Args:
        _data: A list of data collections of the same data type, which will be used
            to color Faces with simulation results. Data collections can be of
            any class (eg. MonthlyCollection, DailyCollection) but they should
            all have headers with metadata dictionaries with 'Surface' keys.
            These keys will be used to match the data in the collections
            to the input faces.
        _hb_objs: An array of honeybee Rooms, Faces, Apertures or Doors to be
            colored with simulation results in the Rhino scene. This can
            also be an entire Model to be colored.
        normalize_: Boolean to note whether results should be normalized by the
            face/sub-face area if the data type of the data_colections supports it.
            If False, values will be generated using sum total of the data collection
            values. Note that this input has no effect if the data type of the
            data_collections is not normalizable since data collection values will
            always be averaged for this case. (Default: True).
        sim_step_: An optional integer (greater than or equal to 0) to select
            a specific step of the data collections for which result values will be
            generated. If None, the geometry will be colored with the total of
            resutls in the data_collections if the data type is cumulative or with
            the average of results if the data type is not cumulative. (Default: None).
        period_: A Ladybug analysis period to be applied to all of the input _data.
        legend_par_: An optional LegendParameter object to change the display
            of the ColorRooms.

    Returns:
        report: ...
        mesh: A colored mesh of the face/sub-face geometry colored using the input
            _data. Multiple meshes will be output for several data collections
            are input.
        wire_frame: A list of polylines representing the outline of the faces.
        legend: Geometry representing the legend for the colored favess.
        title: A text object for the global title.
        faces: A list of honeybee Face, Aperture and Door objects that have been
            matched to the input _data. This can be plugged into the "HB Visualize
            Quick" component to get face breps that are colored.
        colors: A list of color objects that align with the output faces. These
            can be connected to a native Grasshopper "Custom Preview" component
            in order to color room volumes with results.
        values: A list of numbers for each of the faces, which are used to generate
            the colors.
        vis_set: An object containing VisualizationSet arguments for drawing a detailed
            version of the ColorRoom in the Rhino scene. This can be connected to
            the "LB Preview Visualization Set" component to display this version
            of the visualization in Rhino.
"""

ghenv.Component.Name = "HB Color Faces"
ghenv.Component.NickName = 'ColorFaces'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from honeybee.model import Model
    from honeybee.room import Room
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.result.colorobj import ColorFace
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_point3d
    from ladybug_rhino.fromgeometry import from_face3ds_to_colored_mesh, \
        from_face3d_to_wireframe
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.color import color_to_color
    from ladybug_rhino.grasshopper import all_required_inputs
    from ladybug_rhino.config import units_abbreviation
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # extract any faces from input Rooms or Models
    faces = []
    for hb_obj in _hb_objs:
        if isinstance(hb_obj, Model):
            for room in hb_obj.rooms:
                faces.extend(room.faces)
        elif isinstance(hb_obj, Room):
            faces.extend(hb_obj.faces)
        else:
            faces.append(hb_obj)

    # apply analysis period to the data if connected
    if period_ is not None:
        _data = [coll.filter_by_analysis_period(period_) for coll in _data]

    # set default norm_by_floor value
    normalize_ = True if normalize_ is None else normalize_

    # create the ColorFace visualization object and output geometry
    color_obj = ColorFace(_data, faces, legend_par_, sim_step_, normalize_,
                          units_abbreviation())
    graphic = color_obj.graphic_container
    mesh = [from_face3ds_to_colored_mesh([fc], col) for fc, col in
            zip(color_obj.matched_flat_geometry, graphic.value_colors)]
    wire_frame = []
    for face in color_obj.matched_flat_faces:
        wire_frame.extend(from_face3d_to_wireframe(face.geometry))
    legend = legend_objects(graphic.legend)
    title = text_objects(color_obj.title_text, graphic.lower_title_location,
                         graphic.legend_parameters.text_height,
                         graphic.legend_parameters.font)
    faces = color_obj.matched_flat_faces
    colors = [color_to_color(col, 125) for col in graphic.value_colors]
    values = graphic.values
    vis_set = color_obj



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Convert infiltration flow per exterior area measured at a particular blower door pressure
to flow per exterior area measured at a typical building pressure.
_
This is accomplished by computing a normalized-area air mass flow coefficient that
is derived the power law relationship between pressure and air flow.
    Cqa = Qblow / dPblow^n
And then using the coefficient to approximate air flow at typical building pressure.
    Qbldg = Cqa * dPbldg^n
.
where:
    Cqa: Air mass flow coefficient per unit meter at 1 Pa [kg/m2/s/P^n]
    Qblow: Blower-induced volumetric air flow rate per area [m3/s/m2]
    dPblow: Blower-induced change in pressure across building envelope orifice [Pa]
    Qbldg: Typical building volumetric air flow rate per area [m3/s/m2]
    dPbldg: Typical building change in pressure across building envelope orifice [Pa]
    d: Air density [1.2041 kg/m3]
    n: Air mass flow exponent [0.65]
-

    Args:
        _infilt_per_exterior: A numerical value for the intensity of air flow induced by
            blower pressure in m3/s per square meter of exterior surface area.
        _blower_pressure_: A number representing the pressure differential in Pascals (Pa)
            between indoors/outdoors at which the specified _infilt_per_exterior
            occurs. Typical pressures induced by blower doors are 75 Pa and
            50 Pa. (Default: 75).
        _bldg_pressure_: The reference air pressure difference across building envelope
            under typical conditions in Pascals. (Default: 4).

    Returns:
        infilt: The intensity of infiltration in m3/s per square meter of exterior
            surface area at the input _bldg_pressure_.
        C_qa: Air mass flow coefficient per square meter at 1 Pa [kg/m2/s/P^n].
"""

ghenv.Component.Name = 'HB Blower Pressure Converter'
ghenv.Component.NickName = 'BlowerPresure'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from honeybee_energy.properties.room import RoomEnergyProperties
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set default values
    _blower_pressure_ = 75 if _blower_pressure_ is None else _blower_pressure_
    _bldg_pressure_ = 4 if _bldg_pressure_ is None else _bldg_pressure_

    # compute coeffiecient and airflow
    C_qa = RoomEnergyProperties.solve_norm_area_flow_coefficient(
        _infilt_per_exterior, air_density=1, delta_pressure=_blower_pressure_)
    infilt = C_qa * (_bldg_pressure_ ** 0.65)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply process loads to Rooms.
_
Examples of process loads include wood burning fireplaces, kilns, manufacturing
equipment, and various industrial processes. They can also be used to represent 
certain specialized pieces of equipment to be separated from the other end uses,
such as MRI machines, theatrical lighting, elevators, etc.
-

    Args:
        _rooms: Honeybee Rooms to which process loads should be assigned.
        _name_: Text to set the name for the Process load and to be incorporated into a
            unique Process load identifier. If None, a unique name will be
            generated.
        _watts: A number for the process load power in Watts.
        _schedule: A fractional schedule for the use of the process over the course of
            the year. The fractional values will get multiplied by the _watts
            to yield a complete process load profile.
        _fuel_type: Text to denote the type of fuel consumed by the process. Using the
            "None" type indicates that no end uses will be associated with the
            process, only the zone gains. Choose from the following.
                * Electricity
                * NaturalGas
                * Propane
                * FuelOilNo1
                * FuelOilNo2
                * Diesel
                * Gasoline
                * Coal
                * Steam
                * DistrictHeating
                * DistrictCooling
                * OtherFuel1
                * OtherFuel2
                * None
        use_category_: Text to indicate the end-use subcategory, which will identify
            the process load in the EUI output. For example, “Cooking”,
            “Clothes Drying”, etc. (Default: General).
        radiant_fract_: A number between 0 and 1 for the fraction of the total
            process load given off as long wave radiant heat. (Default: 0).
        latent_fract_: A number between 0 and 1 for the fraction of the total
            process load that is latent (as opposed to sensible). (Default: 0).
        lost_fract_: A number between 0 and 1 for the fraction of the total
            process load that is lost outside of the zone and the HVAC system.
            Typically, this is used to represent heat that is exhausted directly
            out of a zone (as you would for a stove). (Default: 0).

    Returns:
        report: Reports, errors, warnings, etc.
        rooms: The input Rooms with process loads assigned to them.
"""

ghenv.Component.Name = 'HB Process Load'
ghenv.Component.NickName = 'Process'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:  # import the honeybee extension
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy extension
    from honeybee_energy.load.process import Process
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, longest_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    rooms = [room.duplicate() for room in _rooms]

    # set default values and check the inputs
    use_category_ = ['Process'] if len(use_category_) == 0 else use_category_
    radiant_fract_ = [0.0] if len(radiant_fract_) == 0 else radiant_fract_
    latent_fract_ = [0.0] if len(latent_fract_) == 0 else latent_fract_
    lost_fract_ = [0.0] if len(lost_fract_) == 0 else lost_fract_
    for i, sched in enumerate(_schedule):
        if isinstance(sched, str):
            _schedule[i] = schedule_by_identifier(sched)

    # loop through the rooms and assign process loads
    for i, room in enumerate(rooms):
        load_watts = longest_list(_watts, i)
        if load_watts != 0:
            name = clean_and_id_ep_string('Process') if len(_name_) == 0 else \
                clean_ep_string(longest_list(_name_, i))
            process = Process(
                '{}..{}'.format(name, room.identifier), load_watts,
                longest_list(_schedule, i), longest_list(_fuel_type, i),
                longest_list(use_category_, i), longest_list(radiant_fract_, i),
                longest_list(latent_fract_, i), longest_list(lost_fract_, i)
            )
            room.properties.energy.add_process_load(process)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a window frame matrial that can be assigned to a Window Construction by
plugging it into the frame_ input of the "HB Window Construction" component.
-

    Args:
        _name: Text to set the name for the frame material and to be incorporated into
            a unique material identifier.
        _width: Number for the width of frame in plane of window [m]. The frame
            width is assumed to be the same on all sides of window.
        _conductance: Number for the thermal conductance of the frame material measured
            from inside to outside of the frame surface (no air films) and taking
            2D conduction effects into account [W/m2-K]. Values for typical frame
            materials are as follows.
                * Aluminum with Thermal Break - 56.4 W/m2-K
                * Aluminum One-Sided (Flush) - 10.7 W/m2-K
                * Wood - 3.5 W/m2-K
                * Vinyl - 2.3 W/m2-K
        _edge_to_cent_: Number between 0 and 4 for the ratio of the glass conductance near
            the frame (excluding air films) divided by the glass conductance
            at the center of the glazing (excluding air films). This is used
            only for multi-pane glazing constructions. This ratio should
            usually be greater than 1.0 since the spacer material that separates
            the glass panes is usually more conductive than the gap between panes.
            A value of 1 effectively indicates no spacer. Values should usually be
            obtained from the LBNL WINDOW program so that the unique characteristics
            of the window construction can be accounted for. (Default: 1).
        outside_proj_: Number for the distance that the frame projects outward from the
            outside face of the glazing [m]. This is used to calculate shadowing
            of frame onto glass, solar absorbed by the frame, IR emitted and
            absorbed by the frame, and convection from frame. (Default: 0).
        inside_proj_: Number for the distance that the frame projects inward from the
            inside face of the glazing [m]. This is used to calculate solar
            absorbed by the frame, IR emitted and absorbed by the frame, and
            convection from frame. (Default: 0).
        _therm_absp_: A number between 0 and 1 for the fraction of incident long
            wavelength radiation that is absorbed by the material. (Default: 0.9).
        _sol_absp_: A number between 0 and 1 for the fraction of incident solar
            radiation absorbed by the material. (Default: 0.7).
        _vis_absp_: A number between 0 and 1 for the fraction of incident
            visible wavelength radiation absorbed by the material.
            Default value is the same as the _sol_absp_.

    Returns:
        frame: A window frame material that can be assigned to a Honeybee
            Window construction.
"""

ghenv.Component.Name = 'HB Window Frame'
ghenv.Component.NickName = 'WindowFrame'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '6'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.material.frame import EnergyWindowFrame
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default material properties
    _edge_to_cent_ = 1 if _edge_to_cent_ is None else _edge_to_cent_
    outside_proj_ = 0 if outside_proj_ is None else outside_proj_
    inside_proj_ = 0 if inside_proj_ is None else inside_proj_
    _therm_absp_ = 0.9 if _therm_absp_ is None else _therm_absp_
    _sol_absp_ = 0.7 if _sol_absp_ is None else _sol_absp_
    name = clean_and_id_ep_string('OpaqueNoMassMaterial') if _name_ is None else \
        clean_ep_string(_name_)

    # create the material
    frame = EnergyWindowFrame(
        name, _width, _conductance, _edge_to_cent_, outside_proj_, inside_proj_,
        _therm_absp_, _sol_absp_, _vis_absp_)
    if _name_ is not None:
        frame.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Spatially visualize the detailed results of a thermal mapping analysis from a
comfort matrix.

-
    Args:
        _comf_mtx: A comfort Matrix object from the "HB Read Thermal Matrix" component.
        _mesh: Mesh objects that correspond with the sensor grids of the thermal map
            analysis. These will be , with a number of faces or vertices that match
            the number of input values and will be colored with results.
        sim_step_: An optional integer (greater than or equal to 0) to select a
            specific time step of the comfort results to be displayed. Note
            that this will override any connected period.
        period_: A Ladybug analysis period to be applied to select a slice of time
            across the comfort results to be displayed.
        legend_par_: An optional LegendParameter object to change the display
            of the results.

    Returns:
        report: ...
        mesh: The input mesh objects colored with results.
        legend: Geometry representing the legend for the results.
        title: A text object for the title.
        colors: The colors associated with each input value.
        values: A list of numbers for each face of the mesh, which are used to
            generate the colors.
"""

ghenv.Component.Name = 'HB Visualize Thermal Map'
ghenv.Component.NickName = 'VizThermalMap'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '7 :: Thermal Map'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug_geometry.geometry3d.mesh import Mesh3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.graphic import GraphicContainer
    from ladybug.legend import LegendParameters
    from ladybug.color import Colorset
    from ladybug.datatype.fraction import RelativeHumidity
    from ladybug.datatype.temperature import Temperature
    from ladybug.datatype.temperaturedelta import TemperatureDelta, RadiantTemperatureDelta
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_mesh3d
    from ladybug_rhino.fromgeometry import from_mesh3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.color import color_to_color
    from ladybug_rhino.grasshopper import all_required_inputs, de_objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def colors_from_data_type(data_type):
    """Get the list of colors that should be used by default for a given data type.

    Args:
        data_type: A data type object that will be used to determine default colors.
    """
    if isinstance(data_type, (Temperature, RadiantTemperatureDelta, RelativeHumidity)):
        return Colorset.original()
    else:  # it is some type of thermal condition or delta temperature
        return Colorset.thermal_comfort()


if all_required_inputs(ghenv.Component):
    # load the data and perform and time-slicing operations on it
    data_mtx = de_objectify_output(_comf_mtx)
    header = data_mtx[0][0].header
    if sim_step_ is not None:
        values = [data[sim_step_] for data_list in data_mtx for data in data_list]
        time_text = data_mtx[0][0].datetimes[sim_step_]
    elif period_ is not None:
        new_data = [[data.filter_by_analysis_period(period_) for data in data_list]
                     for data_list in data_mtx]
        values = [data.average for data_list in new_data for data in data_list]
        time_text = period_
    else:
        values = [data.average for data_list in data_mtx for data in data_list]
        time_text = header.analysis_period

    # generate Ladybug objects for the graphic
    lb_meshes = [to_mesh3d(mesh) for mesh in _mesh]
    lb_mesh = Mesh3D.join_meshes(lb_meshes)
    graphic = GraphicContainer(
        values, lb_mesh.min, lb_mesh.max, legend_par_,
        data_type=header.data_type, unit=header.unit
    )

    # set titles and set default colors and color ranges
    if graphic.legend_parameters.are_colors_default:
        graphic.legend_parameters.colors = colors_from_data_type(header.data_type)
    if isinstance(header.data_type, TemperatureDelta) and not \
            isinstance(header.data_type, RadiantTemperatureDelta) and \
            graphic.legend.is_min_default and graphic.legend.is_max_default:
        graphic.legend_parameters.min = -5
        graphic.legend_parameters.max = 5
    graphic.legend_parameters.title = header.unit
    global_title = '{}\n{}'.format(header.data_type.name, time_text)
    title = text_objects(global_title, graphic.lower_title_location,
                         graphic.legend_parameters.text_height * 1.5,
                         graphic.legend_parameters.font)

    # draw rhino objects
    lb_mesh.colors = graphic.value_colors
    mesh = from_mesh3d(lb_mesh)
    legend = legend_objects(graphic.legend)
    colors = [color_to_color(col) for col in lb_mesh.colors]



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Create a custom ScheduleTypeLimit object that can be assigned to any schedule object.
_
Schedule types exist for the sole purpose of validating schedule values against
upper/lower limits and assigning a data type and units to the schedule values.
As such, they are not necessary to run energy simulations but their use is
generally considered good practice.
-

    Args:
        _name: Text to set the name for the ScheduleTypeLimit. This should be
            unique to avoif conflcit with other schedule types.
        low_limit_: An optional number for the lower limit for values in the
            schedule. If None, there will be no lower limit.
        up_limit_: An optional number for the upper limit for values in the
            schedule. If None, there will be no upper limit.
        discrete_: Boolean to not whether the values of the schedule are continuous
            or discrete. The latter means that only integers are accepted as
            schedule values. Default: False for 'Continuous'.
        _unit_type_: Text for an EnergyPlus unit type, which will be used to assign
            units to the values in the schedule.  Note that this field is not used
            in the actual calculations of EnergyPlus. Default: 'Dimensionless'.
            Choose from the following:
                * Dimensionless
                * Temperature
                * DeltaTemperature
                * PrecipitationRate
                * Angle
                * ConvectionCoefficient
                * ActivityLevel
                * Velocity
                * Capacity
                * Power
                * Availability
                * Percent
                * Control
                * Mode
    
    Returns:
        report: Reports, errors, warnings, etc.
        type_limit: A ScheduleTypeLimit object that can be assigned to any schedule object.
"""

ghenv.Component.Name = "HB Type Limit"
ghenv.Component.NickName = 'TypeLimit'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '2 :: Schedules'
ghenv.Component.AdditionalHelpFromDocStrings = "0"


try:  # import the honeybee-energy dependencies
    from honeybee_energy.schedule.typelimit import ScheduleTypeLimit
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set default values
    numeric_type = 'Discrete' if discrete_ else 'Continuous'
    _unit_type_ = 'Dimensionless' if _unit_type_ is None else _unit_type_

    # create the ScheduleTypeLimit
    type_limit =  ScheduleTypeLimit(
        _name, low_limit_, up_limit_, numeric_type, _unit_type_)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Deconstruct an opaque or window construction into its constituient materials.
-

    Args:
        _constr: An opaque or window construction to be deconstructed. This can
            also be text for an opaque or window construction to be looked
            up in the construction library.

    Returns:
        materials: List of material objects that make up the construction
            (ordered from outside to inside).
        layers: List of material identifiers that make up the construction
            (ordered from outside to inside).
        r_val_si: R-value of the construction in m2-K/W. Note that R-values do NOT
            include the resistance of air films on either side of the construction.
        r_val_ip: R-value of the construction in h-ft2-F/Btu. Note that R-values do NOT
            include the resistance of air films on either side of the construction.
        u_fac_si: U-factor of the construction in W/m2-K.  Note that U-factors
            include the resistance of air films on either side of the construction.
        u_fac_ip: U-factor of the construction in Btu/h-ft2-F.  Note that U-factors
            include the resistance of air films on either side of the construction.
        shgc: The estimated solar heat gain coefficient (SHGC) of the construction.
            This value is produced by finding the solution to the relationship between
            U-value, Solar Transmittance, and SHGC as defined for the simple glazing
            system material in EnergyPlus. More information can be found
            at https://bigladdersoftware.com/epx/docs/9-5/engineering-reference/
            on this partticular sub-page of the engineering reference:
            window-calculation-module.html#step-4.-determine-layer-solar-transmittance
        t_sol: The unshaded shortwave solar transmittance of the construction at normal
            incidence. Note that 'unshaded' in this case means that all blind +
            shade materials in the construction are ignored.
        t_vis: The unshaded visible transmittance of the window at normal incidence.
            Note that 'unshaded' in this case means that all blind + shade
            materials in the construction are ignored.
"""

ghenv.Component.Name = 'HB Deconstruct Construction'
ghenv.Component.NickName = 'DecnstrConstr'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '2'


try:  # import the honeybee-energy dependencies
    from honeybee_energy.construction.opaque import OpaqueConstruction
    from honeybee_energy.construction.window import WindowConstruction
    from honeybee_energy.construction.windowshade import WindowConstructionShade
    from honeybee_energy.construction.dynamic import WindowConstructionDynamic
    from honeybee_energy.lib.constructions import opaque_construction_by_identifier
    from honeybee_energy.lib.constructions import window_construction_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

try:
    from ladybug.datatype.rvalue import RValue
    from ladybug.datatype.uvalue import UValue
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the input
    if isinstance(_constr, str):
        try:
            _constr = opaque_construction_by_identifier(_constr)
        except ValueError:
            _constr = window_construction_by_identifier(_constr)
    else:
        con_cls  = (OpaqueConstruction, WindowConstruction, WindowConstructionShade)
        assert isinstance(_constr, con_cls), \
            'Expected OpaqueConstruction, WindowConstruction, or WindowConstructionShade. ' \
            'Got {}.'.format(type(_constr))

    # get the materials, r-value and u-factor
    materials = _constr.materials
    layers = _constr.layers
    r_val_si = _constr.r_value
    r_val_ip = RValue().to_ip([r_val_si], 'm2-K/W')[0][0]
    u_fac_si = _constr.u_factor
    u_fac_ip = UValue().to_ip([u_fac_si], 'W/m2-K')[0][0]

    # give a warning if there's a negative R-value for a vertical surface
    if r_val_si <= 0:
        msg = 'Construction "{}" has an overall R-value that is less than the\n' \
            'resistance of vertically-oriented air films. This indicates that the ' \
            'construction is only suitable for horizontal/skylight geometry.'.format(
                _constr.display_name)
        print(msg)
        give_warning(ghenv.Component, msg)

    # get the transmittance
    win_types = (WindowConstruction, WindowConstructionShade, WindowConstructionDynamic)
    if isinstance(_constr, win_types):
        shgc = _constr.shgc
        t_sol = _constr.solar_transmittance
        t_vis = _constr.visible_transmittance
    else:
        shgc = 0
        t_sol = 0
        t_vis = 0



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Deconstruct a construction set into its constituient exterior constructions.
-

    Args:
        _constr_set: A construction set to be deconstructed. This can also be
            text for a construction set to be looked up in the construction
            set library.

    Returns:
        exterior_wall: A construction object for the set's exterior walls.
        exterior_roof: A construction object for the set's exterior roofs.
        exposed_floor: A construction object for the set's exposed floors.
        ground_wall: A construction object for the set's underground walls.
        ground_roof: A construction object for the set's underground roofs.
        ground_floor: A construction object for the set's ground-contact floors.
        window: A construction object for apertures with an Outdoors boundary
            condition and a Wall face type for their parent face.
        skylight: A construction object for apertures with an Outdoors boundary
            condition and a RoofCeiling or Floor face type for their parent face.
        operable: A construction object for apertures with an Outdoors boundary
            condition and True is_operable property.
        exterior_door: A construction object for opaque doors with an Outdoors
            boundary condition and a Wall face type for their parent face.
        overhead_door: A construction object for opaque doors with an Outdoors
            boundary condition and a RoofCeiling or Floor face type for their
            parent face.
        glass_door: A construction object for all glass doors with an Outdoors
            boundary condition.
"""

ghenv.Component.Name = "HB Deconstruct ConstructionSet"
ghenv.Component.NickName = 'DecnstrConstrSet'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = "1 :: Constructions"
ghenv.Component.AdditionalHelpFromDocStrings = "2"


try:  # import the honeybee-energy dependencies
    from honeybee_energy.constructionset import ConstructionSet
    from honeybee_energy.lib.constructionsets import construction_set_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))
try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the input
    if isinstance(_constr_set, str):
        _constr_set = construction_set_by_identifier(_constr_set)
    else:
        assert isinstance(_constr_set, ConstructionSet), \
            'Expected ConstructionSet. Got {}.'.format(type(_constr_set))
    
    exterior_wall = _constr_set.wall_set.exterior_construction
    exterior_roof = _constr_set.roof_ceiling_set.exterior_construction
    exposed_floor = _constr_set.floor_set.exterior_construction
    ground_wall = _constr_set.wall_set.ground_construction
    ground_roof = _constr_set.roof_ceiling_set.ground_construction
    ground_floor = _constr_set.floor_set.ground_construction
    window = _constr_set.aperture_set.window_construction
    skylight = _constr_set.aperture_set.skylight_construction
    operable = _constr_set.aperture_set.operable_construction
    exterior_door = _constr_set.door_set.exterior_construction
    overhead_door = _constr_set.door_set.overhead_construction
    glass_door = _constr_set.door_set.exterior_glass_construction



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a simulation output object by selecting sets of commonly-requested output variables.
The resulting object can be used to request output variables from EnergyPlus.
-

    Args:
        zone_energy_use_: Set to True to add outputs for zone energy use when ideal
            air systems are assigned. This includes, ideal air heating + cooling,
            lighting, electric + gas equipment, and fan electric energy.
        hvac_energy_use_: Set to True to add outputs for HVAC energy use from detailed
            systems. This includes outputs for different pieces of equipment,
            which together catch all energy-consuming parts of a system.
            (eg. chillers, boilers, coils, humidifiers, fans, pumps).
        gains_and_losses_: Set to True to Add outputs for zone gains and losses.
            This includes such as people gains, solar gains, infiltration losses/gains,
            and ventilation losses/gains.
        comfort_metrics_: Set to True to add outputs for zone thermal comfort analysis.
            This includes air temperature, mean radiant temperature, relative
            humidity.
        surface_temperature_: Set to True to add outputs for indoor and outdoor
            surface temperature.
        surface_energy_flow_: Set to True to add outputs for energy flow across
            all surfaces.
        load_type_: An optional text value to set the type of load outputs requested.
            Default - 'All'. Choose from the following:
                * All - all energy use including heat lost from the zone
                * Total - the total load added to the zone (both sensible and latent)
                * Sensible - the sensible load added to the zone
                * Latent - the latent load added to the zone
        _report_frequency_: Text for the frequency at which the outputs
                are reported. Default: 'Hourly'.
                Choose from the following:
                    * Annual
                    * Monthly
                    * Daily
                    * Hourly
                    * Timestep
    
    Returns:
        sim_output: A SimulationOutput object that can be connected to the
            "HB Simulation Parameter" component in order to specify which
            types of outputs should be written from EnergyPlus.
"""

ghenv.Component.Name = 'HB Simulation Output'
ghenv.Component.NickName = 'SimOutput'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from honeybee_energy.simulation.output import SimulationOutput
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


# set default reporting frequency.
_report_frequency_ = _report_frequency_.title() \
    if _report_frequency_ is not None else 'Hourly'

# set the default load_type
load_type_ = load_type_.title() if load_type_ is not None else 'All'

# create the starting simulation output object.
sim_output = SimulationOutput(reporting_frequency=_report_frequency_)

# set each of the requested outputs
if zone_energy_use_:
    sim_output.add_zone_energy_use(load_type_)
if hvac_energy_use_:
    sim_output.add_hvac_energy_use()
if gains_and_losses_:
    load_type = load_type_ if load_type_ != 'All' else 'Total'
    sim_output.add_gains_and_losses(load_type)
if comfort_metrics_:
    sim_output.add_comfort_metrics()
if surface_temperature_:
    sim_output.add_surface_temperature()
if surface_energy_flow_:
    sim_output.add_surface_energy_flow()



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Define the window opening properties for all operable apertures of a Room.
_
By default, the properties assigned by this component are translated into simple
ZoneVentilation objects in the resulting IIDF, which can approximate airflow
from both single-sided bouyancy-driven ventilation as well as wind-driven cross
ventilation. Bouyancy-driven flow can happen for essentially all openings while
wind-driven flow can only happen when there are pressure differences across
windows on opposite sides of a Room.
_
Simple ZoneVentilation is computed using the following formulas:
_
VentilationWind = WindCoefficient * OpeningArea * Schedule * WindSpeed
VentilationStack = StackDischargeCoefficient * OpeningArea * ScheduleValue * 
    SQRT(2 * GravityAccelration * HeightNPL * (|(TempZone - TempOutdoors)| / TempZone)) 
TotalVentilation = SQRT((VentilationWind)^2 + (VentilationStack)^2)
_
Note that the (OpeningArea) term is derived from the _fract_area_oper_ and the area
of each aperture while the (HeightNPL) term is derived from the _fract_height_oper_
and the height of each aperture.  The "NPL" stands for "Neutral Plane" and the
whole term represents the height from midpoint of lower opening to the neutral
pressure level of the window (computed as 1/4 of the height of each Aperture in
the translation from honeybee to IDF).
_
More complex airflow phenomena can be modeled by using this component in conjunction
with with the Airflow Network (AFN) component. Note that the window opening
properties assigned by this component are still relevant for such AFN simulations.
-

    Args:
        _rooms: Honeybee Room objects to which window ventilation opening properties
            will be assigned. Note that this component only assigns such properties
            to operable Apertures on the rooms. If the is_operable property
            of any of a room's apertures is not True, no opening properties
            will be assigned.
        _vent_cntrl: A Ventilation Control object from the "HB Ventilation Control"
            component, which dictates the opening behaviour of the Room's apertures.
        _fract_area_oper_: A number between 0.0 and 1.0 for the fraction of the
            window area that is operable. (Default: 0.5, typical of sliding windows).
        _fract_height_oper_: A number between 0.0 and 1.0 for the fraction
            of the distance from the bottom of the window to the top that is
            operable. (Default: 1.0, typical of windows that slide horizontally).
        _discharge_coeff_: A number between 0.0 and 1.0 that will be multipled
            by the area of the window in the stack (buoyancy-driven) part of the
            equation to account for additional friction from window geometry,
            insect screens, etc. (Default: 0.45, for unobstructed windows with
            insect screens). This value should be lowered if windows are of an
            awning or casement type and not allowed to fully open. Some common
            values for this coefficient include the following.
            -
                * 0.0 - Completely discount stack ventilation from the calculation.
                * 0.45 - For unobstructed windows with an insect screen.
                * 0.65 - For unobstructed windows with NO insect screen.
        _wind_cross_vent_: Boolean to indicate if there is an opening of roughly
            equal area on the opposite side of the Room such that wind-driven
            cross ventilation will be induced. If False, the assumption is that
            the operable area is primarily on one side of the Room and there is
            no wind-driven ventilation. If None, the normal vectors of the
            operable aperturs of the input _rooms will be analyzed. If window
            normals of a given room are found to have an angle difference greater
            than 90 degrees, cross ventilation will be set to True. Otherwise,
            it will be False.

    Returns:
        rooms: The input Honeybee Rooms with their window-opening properties edited.
"""

ghenv.Component.Name = 'HB Window Opening'
ghenv.Component.NickName = 'WindowOpen'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:
    from honeybee_energy.ventcool.opening import VentilationOpening
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning, \
        longest_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # loop through the rooms and assign the objects
    op_count = 0
    rooms = []
    for i, room_init in enumerate(_rooms):
        room = room_init.duplicate()  # duplicate to avoid editing the input

        # assign the ventilation control for the windows
        room.properties.energy.window_vent_control = longest_list(_vent_cntrl, i)

        # create the base ventilation opening
        f_area = 0.5 if len(_fract_area_oper_) == 0 else longest_list(_fract_area_oper_, i)
        f_height = 1.0 if len(_fract_height_oper_) == 0 else longest_list(_fract_height_oper_, i)
        discharge = 0.45 if len(_discharge_coeff_) == 0 else longest_list(_discharge_coeff_, i)
        vent_open = VentilationOpening(f_area, f_height, discharge)

        # assign the cross ventilation
        cross_vent = longest_list(_wind_cross_vent_, i) if \
            len(_wind_cross_vent_) != 0 else None
        if cross_vent is None:
            # analyze  normals of room's apertures to test if cross vent is possible
            orient_angles = []
            for face in room.faces:
                for ap in face.apertures:
                    if ap.is_operable:
                        try:
                            orient_angles.append(ap.horizontal_orientation())
                        except ZeroDivisionError:
                            orient_angles.append(0)
            if len(orient_angles) != 0:
                orient_angles.sort()
                vent_open.wind_cross_vent = \
                    True if orient_angles[-1] - orient_angles[0] >= 90 else False
            else:
                vent_open.wind_cross_vent = False
        else:
            vent_open.wind_cross_vent = cross_vent
        vent_aps = room.properties.energy.assign_ventilation_opening(vent_open)
        rooms.append(room)
        op_count += len(vent_aps)

    # give a warning if no operable windows were found among the connected rooms
    if op_count == 0:
        give_warning(
            ghenv.Component, 'No operable Apertures were found among the connected _rooms.\n'
            'Make sure that you have set the is_operable property of Apertures to True.')


# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Set Honeybee Rooms to be conditioned or unconditioned with a heating/cooling system.
_
If _conditioned is True and the connected rooms are not currently conditioned,
an Ideal Air System will be assigned to them. Otherwise, if they are already
conditioned, the existing HVAC system will be left as it is.
-

    Args:
        _rooms: Honeybee Rooms to have their conditioned property set.
        _conditioned: Boolean to indicate whether the rooms are conditioned with a
            heating/cooling system.

    Returns:
        rooms: Rooms that have had their interinal loads removed to reflect a
            plenum space.
"""

ghenv.Component.Name = 'HB Set Conditioned'
ghenv.Component.NickName = 'Conditioned'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '4 :: HVAC'
ghenv.Component.AdditionalHelpFromDocStrings = '1'


try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    rooms = [room.duplicate() for room in _rooms]  # duplicate to avoid editing input
    if _conditioned:
        for room in rooms:
            if not room.properties.energy.is_conditioned:
                room.properties.energy.add_default_ideal_air()
    else:
        for room in rooms:
            room.properties.energy.hvac = None



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Run an OpenStudio Meausre that is intended to create an entire OSM file
(OpenStudio Model). Examples of such measures include the "Create DOE
Prototype Building" measure such as that wich can be downloaded here:
_
https://github.com/NREL/openstudio-model-articulation-gem/tree/develop/lib/
measures/create_DOE_prototype_building

-
    Args:
        _measure: A Measure from the "HB Load Measure" component that is intended to
            generate an OSM from input arguments. Measures can be downloaded
            from the NREL Building Components Library (BCL) at (https://bcl.nrel.gov/).
        add_str_: Optional additional text strings here to be written into the IDF.
            The input here should be complete EnergyPlus objects as a single
            string following the IDF format. This can be used to add addition
            EnergyPlus outputs in the resulting IDF among other features.
        _folder_: An optional folder on this computer, into which the IDF and OSM
            files will be written. If none, a sub-folder within the default
            simulation folder will be used.
        run_: Set to "True" to run the measure and generate the OSM.

    Returns:
        report: Reports, errors, warnings.
        osw: File path to the OpenStudio Workflow JSON on this machine. This workflow
            is executed using the OpenStudio command line interface (CLI) and
            it includes measures to create the OSM from the measure
        osm: The file path to the OpenStudio Model (OSM) that has been generated
            on this computer.
        idf: The file path of the EnergyPlus Input Data File (IDF) that has been
            generated on this computer.
"""

ghenv.Component.Name = 'HB Create OSM Measure'
ghenv.Component.NickName = 'OSMMeasure'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

import sys
import re
import os
import json

try:
    from ladybug.futil import preparedir, nukedir
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.measure import Measure
    from honeybee_energy.run import run_osw
    from honeybee_energy.config import folders as energy_folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from lbt_recipes.version import check_openstudio_version
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # check the presence of openstudio and check that the version is compatible
    check_openstudio_version()

    # process the simulation folder name and the directory
    _folder_ = folders.default_simulation_folder if _folder_ is None else _folder_
    clean_name = re.sub(r'[^.A-Za-z0-9_-]', '_', _measure.display_name)
    directory = os.path.join(_folder_, clean_name, 'openstudio')

    # delete any existing files in the directory and prepare it for simulation
    nukedir(directory, True)
    preparedir(directory)

    # create a dictionary representation of the .osw with steps to run
    # the model measure and the simulation parameter measure
    osw_dict = {'steps': []}
    # assign the measure_paths to the osw_dict
    if 'measure_paths' not in osw_dict:
        osw_dict['measure_paths'] = []
    if energy_folders.honeybee_openstudio_gem_path:  # include honeybee-openstudio measure path
        measure_dir = os.path.join(energy_folders.honeybee_openstudio_gem_path, 'measures')
        osw_dict['measure_paths'].append(measure_dir)

    # add the measure to the OSW
    measure_paths = set()  # set of all unique measure paths
    _measure.validate()  # ensure that all required arguments have values
    measure_paths.add(os.path.dirname(_measure.folder))
    osw_dict['steps'].append(_measure.to_osw_dict())  # add measure to workflow

    # load the inject IDF measure if strings_to_inject have bee specified
    str_inject = None if add_str_ == [] or add_str_[0] is None \
        else '\n'.join(add_str_)
    if str_inject is not None and str_inject != '':
        assert energy_folders.inject_idf_measure_path is not None, \
            'Additional IDF strings input but the inject_idf measure is not installed.'
        idf_measure = Measure(energy_folders.inject_idf_measure_path)
        inject_idf = os.path.join(directory, 'inject.idf')
        with open(inject_idf, "w") as idf_file:
            idf_file.write(str_inject)
        units_arg = idf_measure.arguments[0]
        units_arg.value = inject_idf
        measure_paths.add(os.path.dirname(idf_measure.folder))
        osw_dict['steps'].append(idf_measure.to_osw_dict())  # add measure to workflow

    # write the dictionary to a workflow.osw
    for m_path in measure_paths:
       osw_dict['measure_paths'].append(m_path)
    osw_json = os.path.join(directory, 'workflow.osw')
    if (sys.version_info < (3, 0)):  # we need to manually encode it as UTF-8
        with open(osw_json, 'w') as fp:
            workflow_str = json.dumps(osw_dict, indent=4, ensure_ascii=False)
            fp.write(workflow_str.encode('utf-8'))
    else:
        with open(osw_json, 'wb', encoding='utf-8') as fp:
            workflow_str = json.dump(osw_dict, fp, indent=4, ensure_ascii=False)
    osw = os.path.abspath(osw_json)
    osm, idf = run_osw(osw, silent=False)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Compute spatially-resolved Universal Thermal Climate Index (UTCI) and heat/cold
stress conditions an EPW and Honeybee model.
_
This recipe uses EnergyPlus to obtain surface temperatures and indoor air
temperatures + humidities. Outdoor air temperatures, relative humidities, and
air speeds are taken directly from the EPW. The energy properties of the model
geometry are what determine the outcome of the simulation, though the model's
Radiance sensor grids are what determine where the comfort mapping occurs.
_
Longwave radiant temperatures are obtained by computing spherical view factors
from each sensor to the Room surfaces of the model using Radiance. These view factors
are then multiplied by the surface temperatures output by EnergyPlus to yield
longwave MRT at each sensor. For outdoor sensors, each sensor's sky view is multiplied
by the EPW sky temperature to account for longwave radiant exchange with the sky.
All outdoor context shades and the ground are assumed to be at the EPW air
temperature unless they have been modeled as Honeybee rooms.
_
A Radiance-based enhanced 2-phase method is used for all shortwave MRT calculations,
which precisely represents direct sun by tracing a ray from each sensor to the
solar position. To determine Thermal Comfort Percent (TCP), the occupancy schedules
of the energy model are used for indoor sensors if no schedule_ is input. Any
hour of the energy model occupancy schedule that is 0.1 or greater will be
considered occupied. If no schedule_ is input, all hours of the outdoors are
considered occupied.
-

    Args:
        _model: A Honeybee Model for which UTCI comfort will be mapped. Note that
            this model should have radiance grids assigned to it in order
            to produce meaningful results.
        _epw: Path to an EPW weather file to be used for the comfort map simulation.
        ddy_: Path to a DDY file with design days to be used for the initial sizing
            calculation of the energy simulation. Providing this input is important
            when there are conditioned Room geometries in the model, in which
            case the sizing of the building heating/cooling systems is important
            for modeling the heat exchange between indoors and outdoors.
            Otherwise, it can be ignored with little consequence for
            the simulation.
        north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. This can
            also be Vector for the direction to North. (Default: 0).
        run_period_: An AnalysisPeriod to set the start and end dates of the simulation.
            If None, the simulation will be annual.
        _wind_speed_: A single number for meteorological wind speed in m/s or an hourly
            data collection of wind speeds that align with the input run_period_.
            This will be used for all outdoor comfort evaluation.
            _
            This can also be the path to a folder with csv files that align with
            the model sensor grids. Each csv file should have the same name as
            the sensor grid. Each csv file should contain a matrix of air speed
            values in m/s with one row per sensor and one column per timestep
            of the run period. Note that, when using this type of matrix input,
            these values are not meteorological and should be AT HUMAN SUBJECT LEVEL.
            _
            If unspecified, the EPW wind speed will be used for all outdoor sensors
            and all sensors on the indoors will use a wind speed of 0.5 m/s,
            which is the lowest acceptable value for the UTCI model.
        schedule_: A schedule to specify the relevant times during which comfort
            should be evaluated. This must either be a Ladybug Hourly Data
            Collection that aligns with the input run_period_ or the path to a
            CSV file with a number of rows equal to the length of the run_period_.
            If unspecified, it will be assumed that all times are relevant for
            outdoor sensors and the energy model occupancy schedules will be
            used for indoor sensors.
        comfort_par_: Optional comfort parameters from the "LB UTCI Comfort Parameters"
            component to specify the temperatures (in Celcius) that are
            considered acceptable/comfortable. The default will assume a that 
            the comfort range is between 9C and 26C.
        solar_body_par_: Optional solar body parameters from the "LB Solar Body Parameters"
            object to specify the properties of the human geometry assumed in the
            shortwave MRT calculation. The default assumes average skin/clothing
            absorptivity and a human subject always has their back to the sun
            at a 45-degree angle (SHARP = 135).
        radiance_par_: Text for the radiance parameters to be used for ray
            tracing. (Default: -ab 2 -ad 5000 -lw 2e-05).
        run_settings_: Settings from the "HB Recipe Settings" component that specify
            how the recipe should be run. This can also be a text string of
            recipe settings.
        _run: Set to True to run the recipe and get results. This input can also be
            the integer "2" to run the recipe silently.

    Returns:
        report: Reports, errors, warnings, etc.
        env_conds: A folder containing CSV matrices with all of the environmental conditions
            that were input to the comfort model. These can be loaded into Grasshopper
            using the "HB Read Environment Matrix" component. This includes the following.
                * MRT
                * Air Temperature
                * Relative Humidity
                * Longwave MRT
                * Shortwave MRT Delta
        utci: A folder containing CSV maps of Universal Thermal Climate Index (UTCI)
            temperatures for each sensor grid at each time step of the analysis.
            This can be connected to the "HB Read Thermal Matrix" component to
            parse detailed results into Grasshopper. Values are in Celsius.
        condition: A folder containing CSV maps of comfort conditions for each sensor
            grid at each time step of the analysis. This can be connected to the
            "HB Read Thermal Matrix" component to parse detailed results into
            Grasshopper. Values are as follows.
                * -1 = unacceptably cold conditions
                *  0 = neutral (comfortable) conditions
                * +1 = unacceptably hot conditions
        category: A folder containing CSV maps of the heat/cold stress categories for
            each sensor grid at each time step of the analysis. This can be connected
            to the "HB Read Thermal Matrix" component to parse detailed results
            into Grasshopper. This can be used to understand not just whether
            conditions are acceptable but how uncomfortably hot or cold they
            are. Values indicate the following.
                * -5 = extreme cold stress
                * -4 = very strong cold stress
                * -3 = strong cold stress
                * -2 = moderate cold stress
                * -1 = slight cold stress
                *  0 = no thermal stress
                * +1 = slight heat stress
                * +2 = moderate heat stress
                * +3 = strong heat stress
                * +4 = very strong heat stress
                * +5 = extreme heat stress
        TCP: Lists of values between 0 and 100 for the Thermal Comfort Percent (TCP).
            These can be plugged into the "LB Spatial Heatmap" component along
            with meshes of the sensor grids to visualize spatial thermal comfort.
            TCP is the percentage of occupied time where thermal conditions are
            acceptable/comfortable. Occupied hours are determined from the
            occuppancy schedules of each room (any time where the occupancy
            schedule is >= 0.1 will be considered occupied). Outdoor sensors
            are considered occupied at all times. More custom TCP studies can
            be done by post-processing the condition results.
        HSP: Lists of values between 0 and 100 for the Heat Sensation Percent (HSP).
            These can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize uncomfortably hot locations.
            HSP is the percentage of occupied time where thermal conditions are
            hotter than what is considered acceptable/comfortable. Occupied hours
            are determined from the occuppancy schedules of each room (any time
            where the occupancy schedule is >= 0.1 will be considered occupied).
            Outdoor sensors are considered occupied at all times. More custom HSP
            studies can be done by post-processing the condition results.
        CSP: Lists of values between 0 and 100 for the Cold Sensation Percent (CSP).
            These can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize uncomfortably cold locations.
            CSP is the percentage of occupied time where thermal conditions are
            colder than what is considered acceptable/comfortable. Occupied hours
            are determined from the occuppancy schedules of each room (any time
            where the occupancy schedule is >= 0.1 will be considered occupied).
            Outdoor sensors are considered occupied at all times. More custom CSP
            studies can be done by post-processing the condition results.
"""

ghenv.Component.Name = 'HB UTCI Comfort Map'
ghenv.Component.NickName = 'UTCIMap'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '7 :: Thermal Map'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os

try:
    from lbt_recipes.recipe import Recipe
    from honeybee.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import units_system
    from ladybug_rhino.grasshopper import all_required_inputs, recipe_result
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # create the recipe and set the input arguments
    recipe = Recipe('utci_comfort_map')
    recipe.input_value_by_name('model', _model)
    recipe.input_value_by_name('epw', _epw)
    recipe.input_value_by_name('ddy', ddy_)
    recipe.input_value_by_name('north', north_)
    recipe.input_value_by_name('run-period', run_period_)
    if isinstance(_wind_speed_, str) and os.path.isdir(_wind_speed_):
        recipe.input_value_by_name('air-speed-matrices', _wind_speed_)
    else:
        recipe.input_value_by_name('wind-speed', _wind_speed_)
    recipe.input_value_by_name('schedule', schedule_)
    recipe.input_value_by_name('comfort-parameters', comfort_par_)
    recipe.input_value_by_name('solarcal-parameters', solar_body_par_)
    recipe.input_value_by_name('radiance-parameters', radiance_par_)

    # perform an extra check for units because many people forget to check them
    if isinstance(_model, Model):
        check_model = _model
        if check_model.units != 'Meters':
            check_model = _model.duplicate()
            check_model.convert_to_units('Meters')
        # remove degenerate geometry within native E+ tolerance of 0.01 meters
        for room in check_model.rooms:
            try:
                room.remove_colinear_vertices_envelope(
                    tolerance=0.01, delete_degenerate=True)
            except AssertionError as e:  # room removed; likely wrong units
                error = 'Your Model units system is: {}. ' \
                    'Is this correct?\n{}'.format(_model.units, e)
                raise ValueError(error)

    # run the recipe
    silent = True if _run > 1 else False
    project_folder = recipe.run(
        run_settings_, radiance_check=True, openstudio_check=True, silent=silent)

    # load the results
    try:
        env_conds = recipe_result(recipe.output_value_by_name('environmental-conditions', project_folder))
        utci = recipe_result(recipe.output_value_by_name('utci', project_folder))
        condition = recipe_result(recipe.output_value_by_name('condition', project_folder))
        category = recipe_result(recipe.output_value_by_name('category', project_folder))
        TCP = recipe_result(recipe.output_value_by_name('tcp', project_folder))
        HSP = recipe_result(recipe.output_value_by_name('hsp', project_folder))
        CSP = recipe_result(recipe.output_value_by_name('csp', project_folder))
    except Exception:
        raise Exception(recipe.failure_message(project_folder))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Create a boundary condition representing a custom temperature and/or heat transfer
coefficient on the other side of a face.
-

    Args:
        _temperature_: A temperature value in Celsius to note the temperature on the
            other side of the object. If unspecified, the outdoor air
            temperature will be used.
        heat_coeff_: A value in W/m2-K to indicate the combined convective/radiative film
            coefficient. If equal to 0, then the specified temperature above is
            equal to the exterior surface temperature. Otherwise, the temperature
            above is considered the outside air temperature and this coefficient
            is used to determine the difference between this outside air
            temperature and the exterior surface temperature. (Default: 0).

    Returns:
        bc: A BoundaryCondition object that can be assigned to any honeybee Face object (using
            the "HB Face" component or the "HB Properties by Guide Surface").
"""

ghenv.Component.Name = 'HB Other Side Temperature'
ghenv.Component.NickName = 'OtherTemp'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the core honeybee dependencies
    from honeybee.altnumber import autocalculate
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import honeybee_energy dependencies
    from honeybee_energy.boundarycondition import OtherSideTemperature
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


_temperature_ = _temperature_ if _temperature_ is not None else autocalculate
heat_coeff_ = heat_coeff_ if heat_coeff_ is not None else 0
bc = OtherSideTemperature(_temperature_, heat_coeff_)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create simulation controls with instructions for which types of EnergyPlus
calculations to run.
-

    Args:
        _do_zone_sizing_: Boolean for whether the zone sizing calculation
            should be run. Default: True.
        _do_system_sizing_: Boolean for whether the system sizing calculation
            should be run. Default: True.
        _do_plant_sizing_: Boolean for whether the plant sizing calculation
            should be run. Default: True.
        _for_sizing_period_: Boolean for whether the simulation should
            be run for the sizing periods. Default: False.
        _for_run_period_: Boolean for whether the simulation should
            be run for the run periods. Default: True.
    
    Returns:
        sim_control: A SimulationControl object that can be connected to the
            "HB Simulation Parameter" component in order to specify which
            types of EnergyPlus calculations to run.
"""

ghenv.Component.Name = 'HB Simulation Control'
ghenv.Component.NickName = 'SimControl'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from honeybee_energy.simulation.control import SimulationControl
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


# set default values
_do_zone_sizing_ = _do_zone_sizing_ if _do_zone_sizing_ is not None else True
_do_system_sizing_ = _do_system_sizing_ if _do_system_sizing_ is not None else True
_do_plant_sizing_ = _do_plant_sizing_ if _do_plant_sizing_ is not None else True
_for_sizing_period_ = _for_sizing_period_ if _for_sizing_period_ is not None else False
_for_run_period_ = _for_run_period_ if _for_run_period_ is not None else True

# create the object
sim_control = SimulationControl(_do_zone_sizing_, _do_system_sizing_, _do_plant_sizing_,
                                _for_sizing_period_, _for_run_period_)


# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply a template Service Hot Water (SHW) system to Honeybee Rooms.
_
Note that the rooms must have hot water loads assigned to them in order for them
to be connected to the system.
-

    Args:
        _rooms: Honeybee Rooms to which the input template system will be assigned.
            This can also be a Honeybee Model for which all Rooms will be
            assigned the SHW system.
        _system_type: Text for the specific type of service hot water system and equipment.
            The "HB SHW Templates" component has a full list of the supported
            system templates.
        _name_: Text to set the name for the Service Hot Water system and to be
            incorporated into unique system identifier. If the name is not
            provided, a random name will be assigned.
        _efficiency_: A number for the efficiency of the heater within the system.
            For Gas systems, this is the efficiency of the burner. For HeatPump
            systems, this is the rated COP of the system. For electric systems,
            this should usually be set to 1. If unspecified this value will
            automatically be set based on the equipment_type. See below for
            the default value for each equipment type:
                * Gas_WaterHeater - 0.8
                * Electric_WaterHeater - 1.0
                * HeatPump_WaterHeater - 3.5
                * Gas_TanklessHeater - 0.8
                * Electric_TanklessHeater - 1.0
        _condition_: A number for the ambient temperature in which the hot water tank
            is located [C]. This can also be a Room in which the tank is
            located. (Default: 22).
        _loss_coeff_: A number for the loss of heat from the water heater tank to the
            surrounding ambient conditions [W/K]. (Default: 6 W/K).

    Returns:
        report: Reports, errors, warnings, etc.
        rooms: The input Rooms with a Service Hot Water system applied.
"""

ghenv.Component.Name = "HB SHW System"
ghenv.Component.NickName = 'SHW'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '4 :: HVAC'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

try:  # import the honeybee extension
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
    from honeybee.model import Model
    from honeybee.room import Room
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy extension
    from honeybee_energy.shw import SHWSystem
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # extract any rooms from input Models and duplicate the rooms
    rooms = []
    for hb_obj in _rooms:
        if isinstance(hb_obj, Model):
            rooms.extend([room.duplicate() for room in hb_obj.rooms])
        elif isinstance(hb_obj, Room):
            rooms.append(hb_obj.duplicate())
        else:
            raise ValueError(
                'Expected Honeybee Room or Model. Got {}.'.format(type(hb_obj)))

    # set default value for the inputs
    name = clean_and_id_ep_string('SHW System') if _name_ is None \
        else clean_ep_string(_name_)
    if _condition_ is None:
        _condition_ = 22
    elif isinstance(_condition_, Room):
        _condition_ = _condition_.identifier
    else:
        try:
            _condition_ = float(_condition_)
        except Exception:
            raise ValueError(
                'Input _condition_ must be a Room in which the system is located '
                'or a number\nfor the ambient temperature in which the hot water '
                'tank is located [C].\nGot {}.'.format(type(_condition_))
            )
    _loss_coeff_ = 6 if _loss_coeff_ is None else _loss_coeff_

    # create the SHW System
    shw = SHWSystem(name, _system_type, _efficiency_, _condition_, _loss_coeff_)
    if _name_ is not None:
        shw.display_name = _name_

    # apply the HVAC system to the rooms
    for room in rooms:
        if room.properties.energy.service_hot_water is not None:
            room.properties.energy.shw = shw



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create parameters with criteria for sizing the heating and cooling system.
-

    Args:
        ddy_file_: An optional path to a .ddy file on your system, which contains
            information about the design days used to size the hvac system. If None,
            honeybee will look for a .ddy file next to the .epw and extract all
            99.6% and 0.4% design days.
        filter_ddays_: Boolean to note whether the design days in the ddy_file_
            should be filtered to only include 99.6% and 0.4% design days. If None
            or False, all design days in the ddy_file_ will be incorporated into
            the sizing parameters. This can also be the integer 2 to filter for
            99.0% and 1.0% design days.
        _heating_fac_: A number that will get multiplied by the peak heating load
            for each zone in the model in order to size the heating system for
            the model. Must be greater than 0. (Default: 1.25).
        _cooling_fac_: A number that will get multiplied by the peak cooling load
            for each zone in the model in order to size the cooling system for
            the model. Must be greater than 0. (Default: 1.15).
        eff_standard_: Text to specify the efficiency standard, which will automatically
            set the efficiencies of all HVAC equipment when provided. Note that
            providing a standard here will cause the OpenStudio translation
            process to perform an additional sizing calculation with EnergyPlus,
            which is needed since the default efficiencies of equipment vary
            dependingon their size. THIS WILL SIGNIFICANTLY INCREASE
            TRANSLATION TIME TO OPENSTUDIO. However, it is often
            worthwhile when the goal is to match the HVAC specification with a
             particular standard. The "HB Building Vintages" component has a full
            list of supported HVAC efficiency standards. You can also choose
            from the following.
                * DOE_Ref_Pre_1980
                * DOE_Ref_1980_2004
                * ASHRAE_2004
                * ASHRAE_2007
                * ASHRAE_2010
                * ASHRAE_2013
                * ASHRAE_2016
                * ASHRAE_2019
        climate_zone: Text indicating the ASHRAE climate zone to be used with the
            efficiency_standard. When unspecified, the climate zone will be
            inferred from the design days. This input can be a single
            integer (in which case it is interpreted as A) or it can include the
            A, B, or C qualifier (eg. 3C). Typically, the "LB Import STAT"
            component can yield the climate zone for a particular location.
        building_type: Text for the building type to be used in the efficiency_standard.
            If the type is not recognized or is None, it will be assumed that the
            building is a generic NonResidential. The following have meaning
            for the standard.
                * NonResidential
                * Residential
                * MidriseApartment
                * HighriseApartment
                * LargeOffice
                * MediumOffice
                * SmallOffice
                * Retail
                * StripMall
                * PrimarySchool
                * SecondarySchool
                * SmallHotel
                * LargeHotel
                * Hospital
                * Outpatient
                * Warehouse
                * SuperMarket
                * FullServiceRestaurant
                * QuickServiceRestaurant
                * Laboratory
                * Courthouse

    Returns:
        sizing: Parameters with criteria for sizing the heating and cooling system.
            These can be connected to the "HB Simulation Parameter" component in
            order to specify settings for the EnergyPlus simulation.
"""

ghenv.Component.Name = 'HB Sizing Parameter'
ghenv.Component.NickName = 'SizingPar'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from honeybee_energy.simulation.sizing import SizingParameter
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

# dictionary to get correct efficiency standards
EFF_STANDARDS = {
    'DOE_Ref_Pre_1980': 'DOE_Ref_Pre_1980',
    'DOE_Ref_1980_2004': 'DOE_Ref_1980_2004',
    'ASHRAE_2004': 'ASHRAE_2004',
    'ASHRAE_2007': 'ASHRAE_2007',
    'ASHRAE_2010': 'ASHRAE_2010',
    'ASHRAE_2013': 'ASHRAE_2013',
    'ASHRAE_2016': 'ASHRAE_2016',
    'ASHRAE_2019': 'ASHRAE_2019',
    'pre_1980': 'DOE_Ref_Pre_1980',
    '1980_2004': 'DOE_Ref_1980_2004',
    '2004': 'ASHRAE_2004',
    '2007': 'ASHRAE_2007',
    '2010': 'ASHRAE_2010',
    '2013': 'ASHRAE_2013',
    '2016': 'ASHRAE_2016',
    '2019': 'ASHRAE_2019'
}

# set default sizing factors
heating_fac = 1.25 if _heating_fac_ is None else _heating_fac_
cooling_fac = 1.15 if _cooling_fac_ is None else _cooling_fac_

# create the object
sizing = SizingParameter(None, heating_fac, cooling_fac)

# apply design days from ddy
if ddy_file_ is not None:
    if filter_ddays_ == 1:
        sizing.add_from_ddy_996_004(ddy_file_)
    elif filter_ddays_ == 2:
        sizing.add_from_ddy_990_010(ddy_file_)
    else:
        sizing.add_from_ddy(ddy_file_)

# set the efficiency standard if provided
if eff_standard_ is not None:
    sizing.efficiency_standard = EFF_STANDARDS[eff_standard_]
if climate_zone_ is not None:
    sizing.climate_zone = climate_zone_
if bldg_type_ is not None:
    sizing.building_type = bldg_type_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Create a schedule using from other ScheduleRulesets and AnalysisPeriods over which
each schedule should be applied.
-

    Args:
        _base_schedule: A ScheduleRuleset that represents the base schedule on
            top of which the other _season_scheds will be applied. This can also
            be text to look up a ScheduleRuleset in the schedule library. Any time
            period that is not covered by the input _analysis_periods will default
            to this schedule. Furthermore, the summer and winter design day schedules
            will be taken from this schedule as well as the ScheduleTypeLimits.
        _season_scheds: A list of ScheduleRulesets that align with the _analysis_periods
            below and represent the schedules that will be applied over the
            _base_schedule for the duration of the respective AnalysisPeriod.
            This can also be text to look up ScheduleRulesets in the schedule
            library.
        _analysis_periods: A list of AnalysusPeriod objects that align with the
            _season_scheds and represent the time periods over which each season
            schedule should be applied. Note that, if these AnalysisPeriods
            overlap with one another, then the schedules that come later in
            this list will overwrite those that come earlier in the list for
            the duration of the overlapping time period.
        _summer_des_: An optional list of 24 values that represent the schedule
            values at each hour of the summer design day. This can also be a
            single constant value for the whole day. If None, the summer
            design day schedule of the _base_schedule will be used.
        _winter_des_: An optional list of 24 values that represent the schedule
            values at each hour of the summer design day. This can also be a
            single constant value for the whole day. If None, the summer
            design day schedule of the _base_schedule will be used.
        _name_: Text to set the name for the Schedule and to be incorporated
            into a unique Schedule identifier.

    Returns:
        report: Reports, errors, warnings, etc.
        schedule: A ScheduleRuleset object that can be assigned to a Room, a Load
            object, or a ProgramType object.
        idf_year: Text string for the EnergyPlus ScheduleYear that will ultimately
            be written into the IDF for simulation. This can also be used to add
            the schedule to the schedule library that is loaded up upon the start
            of Honeybee by copying this text into the honeybee/library/schedules/
            user_library.idf file along with the other idf text outputs.
        idf_week: Text string for the EnergyPlus ScheduleWeek that will ultimately
            be written into the IDF for simulation. This can also be used to add
            the schedule to the schedule library that is loaded up upon the start
            of Honeybee by copying this text into the honeybee/library/schedules/
            user_library.idf file along with the other idf text outputs.
        idf_days: Text strings for the EnergyPlus SchedulDays that will ultimately
            be written into the IDF for simulation. This can also be used to add
            the schedule to the schedule library that is loaded up upon the start
            of Honeybee by copying this text into the honeybee/library/schedules/
            user_library.idf file along with the other idf text outputs.
"""

ghenv.Component.Name = "HB Seasonal Schedule"
ghenv.Component.NickName = 'SeasonalSchedule'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '2 :: Schedules'
ghenv.Component.AdditionalHelpFromDocStrings = "4"

try:  # import the core honeybee dependencies
    from ladybug.dt import Date
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.schedule.day import ScheduleDay
    from honeybee_energy.schedule.ruleset import ScheduleRuleset
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check that the input _season_scheds align with the _analysis_periods
    assert len(_season_scheds) == len(_analysis_periods), \
        'Length of the _season_scheds list must match that of the _analysis_periods.' \
        '\n{} does not equal {}'.format(len(_season_scheds), len(_analysis_periods))

    # start by duplicating the base schedule
    name = clean_and_id_ep_string('SeasonalSchedule') if _name_ is None else \
        clean_ep_string(_name_)
    if isinstance(_base_schedule, str):
        _base_schedule = schedule_by_identifier(_base_schedule)
    schedule = _base_schedule.duplicate()
    schedule.identifier = name
    if _name_ is not None:
        schedule.display_name = _name_

    # translate the _season_scheds to individual Rules and apply them to the base
    for season_sch, a_period in zip(_season_scheds, _analysis_periods):
        if isinstance(season_sch, str):
            season_sch = schedule_by_identifier(season_sch)
        if a_period.is_reversed:
            season_rules = season_sch.to_rules(Date(1, 1), a_period.end_time.date) + \
                season_sch.to_rules(a_period.st_time.date, Date(12, 31))
        else:
            season_rules = season_sch.to_rules(
                a_period.st_time.date, a_period.end_time.date)
        for rule in reversed(season_rules):  # preserve priority order of rules
            schedule.add_rule(rule)

    # apply the summer and winter design days if specified
    if len(_summer_des_) != 0:
        s_vals = [_summer_des_[0]] * 24 if len(_summer_des_) == 1 else _summer_des_
        s_id = '{}_SmrDsn'.format(schedule.identifier)
        s_day = ScheduleDay.from_values_at_timestep(s_id, s_vals)
        schedule.summer_designday_schedule = s_day
    if len(_winter_des_) != 0:
        w_vals = [_winter_des_[0]] * 24 if len(_winter_des_) == 1 else _winter_des_
        w_id = '{}_WntrDsn'.format(schedule.identifier)
        w_day = ScheduleDay.from_values_at_timestep(s_id, s_vals)
        schedule.winter_designday_schedule = w_day

    # get the idf strings of the schedule
    idf_year, idf_week = schedule.to_idf()
    idf_days = [day_sch.to_idf(schedule.schedule_type_limit)
                for day_sch in schedule.day_schedules]
